////////////////////////////////////////////////////////////////////////////
// (C) copyright 2011 HC Sharp Software Inc. All Rights Reserved. 
// No duplication or use of this source code is allowed without permission.
// If you use this software you agree to do so at your own risk and you agree
// not to hold the author liable for any damages or injury of any kind. 
// The author assumes NO liability. 
//
// WARNING: This code contains bugs which could cause injury or death
// if you actually used it for any purpose other than your own education. 
/////////////////////////////////////////////////////////////////////////////

// TODO:
// 0) Use one of the remaining analog inputs for a tempurature sensor
// 1) Enable the watchdog timer and monitor for crashes to prevent this
//    thing from competing for headlines with runaway Prius
// 2) Check the timing on ending the pwm signal when the EV releases
// 3) Determine the protocall to re-start charging after a power outage
//    including a random timing delay up to 12 min.
// 4) Implement LCD on RB0:RB3 (low nibble) if you want to waste power on such a thing...

#include <system.h>
#include <icd2.h>		// allocates RAM used by ICD2 debugger

//Set clock frequency
#pragma CLOCK_FREQ	4000000

// duty cycle for 8bit (256 step) scale. <50A d.c.=(A/.6)*2.50  >50A dc=((A/2.5)+64)*2.50
// period = 250 clocks. example 40amp dc=167. Amps = 167/250 * .6 = 40.08amps
#define DUTY16		67
#define DUTY24		100
#define DUTY32		134
#define DUTY40		167
#define DUTY48		200
#define DUTY60		220
#define DUTY70		230

// init eeprom
#pragma DATA _EEPROM, DUTY16, DUTY24, DUTY32, DUTY32, DUTY40, DUTY48, DUTY60, DUTY70

#include "../Include/CtypesEx.h"
#include "TeslaUMC.h"

// Globals
volatile uint Timer1;		// ms countdown timer 



////////////////////////////////////
//***  interrupt routine  **********
////////////////////////////////////
void interrupt( void )
{
	// Handle timer0 interrupt 
	if( intcon & (1<<TMR0IF) )
	{	
		tmr0l = 7;					// preload to make it overflow at 1ms
		if (Timer1) --Timer1;
		//mstimer--;
		//if( !mstimer ) {			// it rolled at 60 seconds
		//	mstimer=60000;
		//	minutes++;
		//}
		clr_bit( intcon, TMR0IF );	// clear timer0 interrupt bit
	}

}

///////////////////////////////////////////////////////////
// read a char from the EEPROM address
///////////////////////////////////////////////////////////
static char readEEProm( unsigned char eeIdx )
{
	eeadr = eeIdx;				// insert address into EE address register (0-255)
	clr_bit( eecon1, EEPGD );	// access EEPROM memory (not program memory) 	
	clr_bit( eecon1, CFGS );	// don't access configuration ROM
	set_bit( eecon1, RD );		// start the read operation
	return eedata;
} 

///////////////////////////////////////////////////////////
// write a char to EEPROM address
///////////////////////////////////////////////////////////
static void writeEEProm( unsigned char eeIdx, char wdata )
{
	eeadr = eeIdx;				// insert address into EE address register (0-255)
	eedata = wdata;
	clr_bit( eecon1, EEPGD );	// access EEPROM memory (not program memory) 	
	clr_bit( eecon1, CFGS );	// don't access configuration ROM
	set_bit( eecon1, WREN );	// enable writes
	clr_bit( intcon, GIE );		// disable interrupts. Avg write takes 4ms!!! 
	asm movlw 0x55				// required sequence for eeprom write from datasheet
	asm movwf _eecon2
	asm movlw 0x0AA
	asm movwf _eecon2
	set_bit( eecon1, WR );		// start the write operation
	set_bit( intcon, GIE );		// re-enable interrupts 
	while( test_bit(eecon1,WR) )// wait for slooooow write to finish
		;
	clr_bit( eecon1, WREN );	// disable writes
} 




////////////////////////////////////
//***  main  ***********************
////////////////////////////////////
void main()
{
	uchar dutyCycle = 0;
	uchar a, show_life = 4, evResponse = 0;
	uchar ledPort = LED_XA_PORT;
	uchar ledPin = LED_16A;
	uchar tmr2_lo, tmr2_hi;		// for testing
	uint  avgAD = 65000;
	uchar temp;

	
	// *** SETUP ***
	// set external osc mode
	clr_bit( osccon, SCS1 );	// 00 = external primary oscillator (default);	Use config HS mode
	clr_bit( osccon, SCS0 );
	
	Reset:
	// A/D 
	adcon1 = 0b00001110;		// analog control register: 1=Digital I/O	an3-an0
	set_bit( adcon2, ADFM  );	// right justify result (6 msb are 0)
	clr_bit( adcon2, ACQT2 );	// Acq time (Tacq)		111=20Tad; 110=16Tad; 101=12Tad; 100=8Tad; 
	clr_bit( adcon2, ACQT1 );	// 						011=6Tad; 010=4Tad; 001=2Tad; 000=0Tad
	set_bit( adcon2, ACQT0 );
	set_bit( adcon2, ADCS2 );	// 100=Fosc/4 for up to 5.71 Mhz. 
	clr_bit( adcon2, ADCS1 );
	clr_bit( adcon2, ADCS0 );
	set_bit( adcon0, ADON );	// enable A/D converter

	// Ports I/O
	porta = 0b00000110;			// 
	trisa = 0b00000001;			// EV volts input. others are output
	portb = 0b00000000;
	trisb = 0b11100000;			// output=0	except RB7-RB6 programming inputs, RB5 button
	portc = 0b11111111;			// all LEDs need to be hi to be off
	trisc = 0b00000000;			// output=0	all outputs
	porte = 0b00001000;
	trise = 0b00001000;			// output=0	all inputs
	
	// timer0
	//set_bit( t0con, T08BIT );	// set as 8 bit timer	(default=1)
	clr_bit( t0con, T0CS );		// driven by Fosc/4 (not by pin)
	clr_bit( t0con, PSA ); 		// 0 = prescaler is assigned
	clr_bit( t0con, T0PS2 );	// 001=prescaler 1:4	
	clr_bit( t0con, T0PS1 );	// at 4MHz 1:4 = 250 counts per ms
	set_bit( t0con, T0PS0 );	
	set_bit( t0con, TMR0ON); 	// turn it on

	// timer2
	set_bit( t2con, TMR2ON );	// enable tmr2
	clr_bit( t2con, T2CKPS1 );	// 01=prescaler 1:4	
	set_bit( t2con, T2CKPS0 );	// at 4MHz 1:4 = 250 counts per ms

	// interrupts
	//set_bit(intcon, PEIE);	// enable peripheral interrupts. (tmr0 is not a peripheral, but tmr1 is)
	set_bit(intcon, GIE);		// enable (or disable =0) general interrupts
	clr_bit(intcon, TMR0IF);	// clear interrupt flag before enableing
	set_bit(intcon, TMR0IE);	// 1= enable timer0 overflow int
	
	clr_bit(pir1, ADIF);		// AD int. clear before enabling
	set_bit(pie1, ADIE);		// Enables the A/D converter int
	
	// PWM 
	clr_bit(ccp1con, DC1B1);	// lsb of 10 bit duty cycle
	clr_bit(ccp1con, DC1B0);	// lsb of 10 bit duty cycle
	ccpr1l = 255;				// duty cycle 8 msb; set higher than PR2 (PWM period) and PWM output pin is set
	clr_bit(ccp1con, CCP1M3);	// set CCP1 into PWM mode: 11xx
	clr_bit(ccp1con, CCP1M2);	// disable it for now (we don't want to start the pwm just yet)
	
	// turn out the lights
	set_bit(LED_GRN_PORT, LED_GRN_PIN);	// hi = off
	set_bit(LED_YEL_PORT, LED_YEL_PIN);
	set_bit(PWM_PORT, PWM_PIN);
	portc = 0xFF;				// shortcut instead of setting each pin which would be the right way

	//*** end setup ***	

	
	
	// show some life. Only do this once. Will be skipped in reset loop cuz show_life will be 0
	for( ; show_life; show_life--) {
		clr_bit(LED_GRN_PORT, LED_GRN_PIN);	// lo = on
		clr_bit(LED_YEL_PORT, LED_YEL_PIN);
		delay_ms(100);
		set_bit(LED_GRN_PORT, LED_GRN_PIN);	// hi = off
		set_bit(LED_YEL_PORT, LED_YEL_PIN);
		delay_ms(100);
	}
	
	// State A	(12v solid)
	set_bit(PWM_PORT, PWM_PIN);

	// get last PWM duty cycle
	dutyCycle = DUTY40;					// default
	ledPin = LED_40A;					// default
	
	ledPin = readEEProm( 8 );			// read from eeprom here (eeprom addr 8). if that fails, use default 40A
	if( ledPin > 7 )
		ledPin = LED_40A;
	dutyCycle = readEEProm( ledPin );	// get the duty cycle for this amp setting	
	clr_bit( LED_XA_PORT, ledPin );		// light the amp indicator led
	
	
	// check for 9v (EV present, but not ready for power)
	// Stage A = 9v = vehicle present, not ready for power.		A/D = 52,100 
	// Stage B = 9v Start pwm to let EV know what's available.	A/D = 39,600
	// Stage C = 6v EV about to begin drawing power.			A/D = 26,100
	// Stage ? = 3v EV requires ventilation which we don't have.
	evResponse = 0;
	while( evResponse < 40 ) {		// Ensure 40 cycles to avoid glitches while waiting for drop to 9v
		avgAD = getAD();
		if( avgAD < 46000 ) {
			clr_bit(LED_GRN_PORT, LED_GRN_PIN);
			evResponse += 2;
		}
		else {
			set_bit(LED_GRN_PORT, LED_GRN_PIN);
			if( evResponse > 0 )
				evResponse -= 1;
		}
		
		// check for button pressed to change duty cycle
		if( test_bit( BUTTON_PORT, BUTTON_PIN ) == 0 ) {
			ledPin++;							// increment to next pin/led/amp setting
			if( ledPin == CCP1 )					// skip RC2 pin which is not an LED
				ledPin++;
			if( ledPin > RC7 )
				ledPin = RC0;
				
			// read eeprom index ledPin 
			dutyCycle = readEEProm( ledPin );	// get the duty cycle for this amp setting
			
			// light correct led
			portc = 0xFF;
			clr_bit( LED_XA_PORT, ledPin );	
				
			// save this setting to eeprom
			writeEEProm( 8, ledPin );
			
			// release button
			while( test_bit( BUTTON_PORT, BUTTON_PIN ) == 0 )
				delay_ms(5);
		}
		
		delay_ms(10);
	}
	
	
	// Start PWM
	set_bit(ccp1con, CCP1M3);	// set CCP1 into PWM mode: 11xx
	set_bit(ccp1con, CCP1M2);
	pr2 = 249;
	ccpr1l = dutyCycle;

	// monitor line to make sure EV is still hooked up
	evResponse = 0;
	while( evResponse < 200 ) {
		// now that PWM is on, we can only check the EV response voltage on the high PWM cycle
		// and if the duty cycle is very low, like 12 amps = 20%, we have a .2ms window, to get A/D.
		// that is only 200 instruction cycles. It appears to be enough, but the op-amp takes 10+ tmr2 clocks
		// of fall time before it can be measured reliably. We use 20 to be safe. 
		
		while( test_bit( PWM_PORT, PWM_PIN ) ) 	// wait for it to go low
			tmr2_lo = tmr2;
		while( !test_bit( PWM_PORT, PWM_PIN ) )	// wait for it to go hi
			tmr2_hi = tmr2;
		if( dutyCycle > 40 ) {					// pretty much everything. can probably take this test out.
			while( tmr2 < 20 )					// wait for op-amp to fall. should only need about 10.
				tmr2_hi = tmr2;
			// start A/D
			set_bit( adcon0, GO );
			while( adcon0 & (1<<DONE) )
				;
			avgAD = adresl + (adresh << 8);		// should try avgAD = adres
			avgAD = avgAD << 6;					// normally we take 64 samples, so multiply by 64 to get comparable result.
		}
		else
			avgAD = 45000;
		
		if( avgAD > 46000 ) {
			evResponse += 1;
		}
		else {
			if( evResponse > 10 )
				evResponse -= 10;
		}
		
		if( !Timer1 ) {
			Timer1 = 500;
			toggle_bit( LED_YEL_PORT, LED_YEL_PIN );
			toggle_bit( LED_XA_PORT, ledPin );
		}
		
		delay_ms(10);
	}
	
	// turn off the lights
	set_bit( LED_GRN_PORT, LED_GRN_PIN );
	set_bit( LED_YEL_PORT, LED_YEL_PIN );
	set_bit( LED_XA_PORT, ledPin);
	set_bit( PWM_PORT, PWM_PIN );
	ccpr1l = 255;					// latch it hi. If it goes low for more than 1ms then the EV thinks its disconnected.

	goto Reset;
		
}



////////////////////////////////////////////////////////////////////
// Get AD
////////////////////////////////////////////////////////////////////
// samples AD pin 64 times and returns average in 16 bits
uint getAD()
{
	uchar a;
	uint  avgAD;
	twoBytesUnion adSampleRes;
	
	a=64;			// counts the number of samples
	avgAD = 0;		// total of 64 samples
	adresh = 0;		// ad registers
	adresl = 0;
	adSampleRes.uWord = 0;
	while( a ) {
		// start AD by setting GO
		set_bit( adcon0, GO );
		adSampleRes.bytes.cLoByte = adresl;		// kinda risky hoping this finishes before conversion starts
		adSampleRes.bytes.cHiByte = adresh;
		avgAD += adSampleRes.uWord;				// add this here while we're doing nothing waiting for acq and conversion
		while( adcon0 & (1<<DONE) )
			;
		a--;
	}
	adSampleRes.bytes.cLoByte = adresl;			// get the last one!
	adSampleRes.bytes.cHiByte = adresh;
	avgAD += adSampleRes.uWord;					// last one
	
	return avgAD;
}



